/* Stylesheet for jupyter-sphinx

These styles mimic the Jupyter HTML styles.

The default CSS (Cascading Style Sheet) class structure of jupyter-sphinx
is the following:

jupyter_container
  code_cell (optional)
  stderr (optional)
  output (optional)

If the code_cell is not displayed, then there is not a jupyter_container, and
the output is provided without CSS.

This stylesheet attempts to override the defaults of all packaged Sphinx themes
to display jupter-sphinx cells in a Jupyter-like style.

If you want to adjust the styles, add additional custom CSS to override these
styles.

After a build, this stylesheet is loaded from ./_static/jupyter-sphinx.css .

*/


div.jupyter_container {
    padding: .4em;
    margin: 0 0 .4em 0;
    background-color: #FFFF;
    border: 1px solid #CCC;
    -moz-box-shadow: 2px 2px 4px rgba(87, 87, 87, 0.2);
    -webkit-box-shadow: 2px 2px 4px rgba(87, 87, 87, 0.2);
    box-shadow: 2px 2px 4px rgba(87, 87, 87, 0.2);
}
.jupyter_container div.code_cell {
  border: 1px solid #cfcfcf;
  border-radius: 2px;
  background-color: #f7f7f7;
  margin: 0 0;
  overflow: auto;
}

.jupyter_container div.code_cell pre {
  padding: 4px;
  margin: 0 0;
  background-color: #f7f7f7;
  border: none;
  background: none;
  box-shadow: none;
  -webkit-box-shadow: none; /* for nature */
  -moz-box-shadow: none; /* for nature */
}

.jupyter_container div.code_cell * {
  margin: 0 0;
}
div.jupyter_container div.highlight {
  background-color: #f7f7f7; /* for haiku */
}
div.jupyter_container {
    padding: 0;
    margin: 0;
}

/* Prevent alabaster breaking highlight alignment */
div.jupyter_container .hll {
  padding: 0;
  margin: 0;
}

/* overrides for sphinx_rtd_theme */
.rst-content .jupyter_container div[class^='highlight'],
.document .jupyter_container div[class^='highlight'],
.rst-content .jupyter_container pre.literal-block {
  border:none;
  margin: 0;
  padding: 0;
  background: none;
  padding: 3px;
  background-color: transparent;
}
/* restore Mathjax CSS, as it assumes a vertical margin. */
.jupyter_container .MathJax_Display {
  margin: 1em 0em;
  text-align: center;
}
.jupyter_container .stderr {
    background-color: #FCC;
    border: none;
    padding: 3px;
}
.jupyter_container .output {
    border: none;
}
.jupyter_container div.output pre {
    background-color: white;
    background: none;
    padding: 4px;
    border: none;
    box-shadow: none;
    -webkit-box-shadow: none; /* for nature */
    -moz-box-shadow: none; /* for nature */
}
.jupyter_container .code_cell td.linenos {
  text-align: right;
  padding: 4px 4px 4px 8px;
  border-right: 1px solid #cfcfcf;
  color: #999;
}
.jupyter_container .output .highlight {
  background-color: #ffffff;
}
/* combine sequential jupyter cells,
   by moving sequential ones up higher on y-axis */
div.jupyter_container + div.jupyter_container {
    margin: -.5em 0 .4em 0;
}

/* Fix for sphinx_rtd_theme spacing after jupyter_container #91 */
.rst-content .jupyter_container {
    margin: 0 0 24px 0;
}
